#pragma once

#include "memory.h"

// Screen dimensions
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64

// VRAM Size & Address
// 1024 for 128*64 monochrome
// + 1 to prevent out of bounds access on Gravity Duck
#define VRAM_SIZE (SCREEN_WIDTH * SCREEN_HEIGHT / 8 + 1)    
// Arbitrary address relative to start of RAM
// + needs to be un-aligned (+1) for correct behavior with MonochromeLib
#define VRAM_ADDRESS (RAM_ADDRESS_P1 + RAM_SIZE + 2048 + 1) 
                                                          
// LCD registers addresses
#define LCD_SELECT_REGISTER 0xB4000000
#define LCD_DATA_REGISTER   0xB4010000

// Character Sets support
#define CHARACTER_SET_SIZE  (4 * 256 * 7) // 4 * 256 characters, 8 * 7 pixels packed
#define CHAR_WIDTH   6
#define CHAR_HEIGHT  8

// FKeyIcon support
#define FKEYICON_ADDRESS    0x8004CCF2
#define FKEYICON_SIZE       (1017 * 24) // 1017 fkey icons (F1-F6), 24 * 8 pixels packed
#define FKEYICON_WIDTH      24
#define FKEYICON_HEIGHT     8

// OpCodes support
#define OPCODES_SIZE        (1236 * 15) // 1236 operation codes, 0-15 characters

// Cursor icons
#define CURSOR_ICONS_SIZE   (10 * 7)   // 10 cursor icons, 8 * 7 pixels packed
#define CURSOR_ICONS_WIDTH  6
#define CURSOR_ICONS_HEIGHT 7

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_Cursor.HTM
typedef struct {
    int8_t  col; // starts at 1
    int8_t  row; // starts at 1
    int16_t flash_style; // flashing icon style
    int8_t  flash_mode;  // 0: no flashing cursor, 1: flashing cursor
    int8_t  blink_state; // (used when cursor is flashing) 0: not visible, 1: visible
} cursor_t;

struct display_t {
    // Pointer to virtual memory in RAM
    uint8_t vram[VRAM_SIZE];

    // LCD screen emulation
    uint8_t lcd[VRAM_SIZE]; // LCD pixel data
    uint8_t lcd_registers;  // Use for both selector & data registers

    // Saved display images
    uint8_t saved_disps[VRAM_SIZE * 3]; // used in SaveDisp and RestoreDisp syscalls

    // Locate position
    uint32_t loc_x;
    uint32_t loc_y;

    // Cursor informations
    cursor_t cursor;

    // Character set bitmaps
    uint8_t character_set[CHARACTER_SET_SIZE];
    uint8_t character_set_mini[CHARACTER_SET_SIZE];

    // FKeyIcons bitmaps
    uint8_t fkey_icons[FKEYICON_SIZE];

    // OpCodes strings
    uint8_t opcodes[OPCODES_SIZE];

    // Cursor bitmaps
    uint8_t cursor_icons[CURSOR_ICONS_SIZE];
};