﻿#pragma once
#pragma warning (disable : 6011 6387 28182)

typedef struct memory_t memory_t;
typedef struct cpu_t cpu_t;
typedef struct display_t display_t;
typedef struct keyboard_t keyboard_t;
typedef struct fs_t fs_t;
typedef struct malloc_manager_t malloc_manager_t;

#define USE_EMSCRIPT

#ifdef USE_EMSCRIPT
#include <emscripten.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include "memory.h"
#include "display.h"
#include "utils.h"
#include "instructions/instructions.h"
#include "syscalls/syscalls.h"

// Initialization
#define PC_PROGRAM_START 0x00300200 // Execution starting address
#define PR_INIT_VALUE    0xffeeeeee // Random value used to check for app return
#define SR_INIT_VALUE    0x700000f0 // Status Register initialization value (from doc)

#define SYSCALL_ADDRESS  0x80010070 // The calling convention to access the system calls, is to jump to 0x80010070 with the syscall number in r0

// CPU
struct cpu_t {
    // General registers R0 - R15
    int32_t r[16];

    // Control registers
    uint32_t gbr;                   // Global base register
    uint32_t sr;                    // Status register
    uint32_t ssr;                   // Saved status register
    uint32_t spc;                   // Saved program counter
    uint32_t vbr;                   // Vector base register
    uint32_t sgr;                   // Saved general register
    uint32_t dbr;                   // Debug base register

    // System registers
    uint32_t pc;                    // Program counter
    uint32_t pr;                    // Procedure register
    uint32_t mach;                  // Multiply-accumulate high
    uint32_t macl;                  // Multiply-accumulate low

    // Debug
    uint32_t instruction_count;     // Number of instructions executed
    uint32_t instruction_per_frame; // Number of instructions to execute each frame
    uint32_t execution_paused;      // Whether the execution is paused
    uint32_t execution_finished;    // Whether the program has reached the end
    uint32_t execute_one_step;      // Whether to execute only one step
    uint32_t next_breakpoint;       // The execution will pause when reaching this address

    // External Components
    memory_t* mem;                  // Memory
    display_t* disp;                // Display
    keyboard_t* keyboard;           // Keyboard

    fs_t* fs;                         // Custom file system
    malloc_manager_t* malloc_manager; // malloc/realloc/free manager

    int32_t RTC_Ticks;              // 1/128 second ticks (updated from javascript)
};

#define set_bit(value, bit) ((value) |= (1 << (bit)))
#define clear_bit(value, bit) ((value) &= ~(1 << (bit)))
#define get_bit(value, bit) (((value) >> (bit)) & 1)
#define toggle_bit(value, bit) ((value) ^= (1 << (bit)))
#define update_bit(value, bit, new_bit) ((value) = ((value) & ~(1 << (bit))) | ((new_bit) << (bit)))

void run_next_instruction(cpu_t* cpu);

void main_loop_html(void* arg);
void init_loop_html(cpu_t* cpu);
void init_loop_c(cpu_t* cpu);