#pragma once

#include "main.h"

// Memory start addresses
#define ROM_ADDRESS        0x00300000
#define RAM_ADDRESS_P0     0x08100000
#define RAM_ADDRESS_P1     0x88000000 // 0x8801c000 // 0x88048000 on 35+E II
#define RAM_ADDRESS_P2     0xA8000000
#define XRAM_ADDRESS       0xE5007000
#define YRAM_ADDRESS       0xE5017000
#define ILRAM_SADDRESS     0xE5200000

// Memory sizes
#define ROM_SIZE         (2048 * 1024)
#define RAM_SIZE         (1024 * 1024)
#define XRAM_SIZE           (8 * 1024)
#define YRAM_SIZE           (8 * 1024)
#define ILRAM_SIZE          (4 * 1024)

typedef struct{
    uint32_t size;
    uint32_t addr;
} malloc_info_t;

// MEMORY
struct memory_t {
    uint8_t* rom;          // Read-only memory
    uint32_t rom_size;

    uint8_t ram[RAM_SIZE]; // Random-access memory

    uint8_t* tmp;

    // uint8_t xram[XRAM_SIZE];
    // uint8_t yram[YRAM_SIZE];
    // uint8_t ilram[ILRAM_SIZE];

    // Malloc
    int mallocCount;
    malloc_info_t* mallocs;
};

uint8_t* get_memory_for_address(cpu_t* cpu, uint32_t address);

uint32_t mem_read(cpu_t* cpu, uint32_t address, uint8_t bytes);
void mem_write(cpu_t* cpu, uint32_t address, uint32_t data, uint8_t bytes);

uint32_t emulate_keyboard_register_read(cpu_t* cpu);
void emulate_lcd_register_write(cpu_t* cpu, uint32_t address, uint32_t data);