#pragma once

#include "syscalls.h"

// Bfile open modes
#define OPENMODE_READ            0x01
#define OPENMODE_READ_SHARE      0x80
#define OPENMODE_WRITE           0x02
#define OPENMODE_READWRITE       0x03
#define OPENMODE_READWRITE_SHARE 0x83

// FILE_INFO.type
#define DT_DIRECTORY              0x0000      // Directory
#define DT_FILE                   0x0001      // File
#define DT_ADDIN_APP              0x0002      // Add-In application
#define DT_EACT                   0x0003      // eActivity
#define DT_LANGUAGE               0x0004      // Language
#define DT_BITMAP                 0x0005      // Bitmap
#define DT_MAINMEM                0x0006      // Main Memory data
#define DT_TEMP                   0x0007      // Temporary data
#define DT_DOT                    0x0008      // .  (Current directory)
#define DT_DOTDOT                 0x0009      // .. (Parent directory)
#define DT_VOLUME                 0x000A      // Volume label

// Informations about a file used in Bfile functions
typedef struct {
    char name[40];     // File name
    FILE* file_handle; // File handle (fopen)
    int32_t handle;    // File handle (SDK)
    int32_t seek_pos;  // Current seek position
    int32_t size;      // File size
} file_t;

// Informations about a search used in Bfile_FindFirst/Next functions
typedef struct {
    char pathname[40]; // Search value (e.g. "*.g1a")
    int32_t index;     // Current index in the search
    int32_t handle;    // Search handle
} file_search_t;

struct fs_t {
    // Open files (Only 4 files can be opened at the same time on real hardware)
    file_t files[4];
    uint8_t open_files_count;

    // Search handles
    file_search_t searches[4];
    uint8_t search_count;

    // MCS Handle
    const char* mcs_filename;
    int32_t mcs_filesize;

    uint8_t needs_sync; // Tells javascript to persist the files to the local storage
};

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_Bfile.htm
void syscall_Bfile_CreateEntry(cpu_t* cpu, uint32_t filename_ptr, int mode, uint32_t size_ptr);
void syscall_Bfile_DeleteEntry(cpu_t* cpu, uint32_t filename_ptr, int mode);
void syscall_Bfile_FindFirst(cpu_t* cpu, uint32_t pathname_ptr, uint32_t findhandle_ptr, uint32_t foundfile_ptr, uint32_t fileinfo_ptr);
void syscall_Bfile_FindNext(cpu_t* cpu, uint32_t findhandle, uint32_t foundfile_ptr, uint32_t fileinfo_ptr);
void syscall_Bfile_FindClose(cpu_t* cpu, uint32_t findhandle);
void syscall_Bfile_OpenFile(cpu_t* cpu, uint32_t filename_ptr, int mode, int mode2);
void syscall_Bfile_CloseFile(cpu_t* cpu, int handle);
void syscall_Bfile_WriteFile(cpu_t* cpu, int handle, uint32_t buf_ptr, int size);
void syscall_Bfile_ReadFile(cpu_t* cpu, int handle, uint32_t buf_ptr, int size, int readpos);
void syscall_Bfile_SeekFile(cpu_t* cpu, int handle, int pos);
void syscall_Bfile_GetFileSize(cpu_t* cpu, uint32_t handle);
void syscall_Bfile_Bfile_GetMediaFree(cpu_t* cpu, uint32_t media_id_ptr, uint32_t freespace_ptr);

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_MCShandles.HTM
void syscall_MCSGetDlen2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t len_ptr);
void syscall_MCSGetData1(cpu_t* cpu, int32_t offset, int32_t len_to_copy, int32_t buffer_ptr);
void syscall_MCSPutVar2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t data_len, int32_t buffer_ptr);
void syscall_MCSOvwDat2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t bytes_to_write, int32_t buffer_ptr, int32_t write_offset);
void syscall_MCSDelVar2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr);

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_MCS.htm
void syscall_MCS_GetState(cpu_t* cpu, int32_t maxspace_ptr, int32_t currentload_ptr, int32_t remainingspace_ptr);
void syscall_MCS_SearchDirectoryItem(cpu_t* cpu, int32_t dir_ptr, int32_t item_name_ptr, int32_t flags_ptr, int32_t item_ptr, int32_t data_ptr, int32_t len_ptr);
void syscall_MCS_OverwriteData(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t write_offset, int32_t bytes_to_write, int32_t buffer_ptr);
void syscall_MCS_DeleteItem(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr);