#pragma once

#include "syscalls.h"

// Keyboard register addresses (SH3-specific)
#define KB_PORTB_CTRL        0xA4000102
#define KB_PORTM_CTRL        0xA4000118
#define KB_PORTA             0xA4000120
#define KB_PORTB             0xA4000122
#define KB_PORTM             0xA4000138
#define KB_SIZE             (0xA4000138 - 0xA4000102 + 1) // Max - min address + 1 (insclusive)
#define KB_ADDRESS             KB_PORTB_CTRL

// Keyboard register addresses (SH4-specific)
#define KEYSC_ADDRESS        0xA44B0000
#define KEYSC_SIZE           (6 * 2)

// GetKey codes
#define KEY_EXIT  30002
#define KEY_EXE   30004
#define KEY_SHIFT 30006
#define KEY_ALPHA 30007
#define KEY_AC    30015

// Repeat key default values
#define DEFAULT_REPEAT_TIME_FIRST_COUNT 20
#define DEFAULT_REPEAT_TIME_NEXT_COUNT  5

struct keyboard_t {
    uint8_t row_state[10]; // Keyboard rows state
    uint8_t keyboard_registers[KB_SIZE]; // Keyboard registers

    // Buffered keys that are waiting to be read
    uint16_t key_buffer[16];
    uint32_t buffered_keys_count;

    // Shift/Alpha mode states 
    uint8_t shift_mode;
    uint8_t alpha_mode;

    // Control key repeat delays, used in Bkey syscalls
    uint32_t repeat_time_first;
    uint32_t repeat_time_next;
};

void syscall_GetKey(cpu_t* cpu, unsigned int keycode_address);
void syscall_KeyBoard_GetKeyWait(cpu_t* cpu, uint32_t column_ptr, uint32_t row_ptr, uint32_t type_of_waiting, uint32_t timeout_period, uint32_t menu, uint32_t keycode_ptr);

void syscall_PutKey(cpu_t* cpu, uint32_t keycode, uint32_t mode);
void syscall_Keyboard_ClrBuffer(cpu_t* cpu);

void syscall_Bkey_Set_RepeatTime(cpu_t* cpu, int32_t first_count, int32_t next_count);
void syscall_Bkey_Get_RepeatTime(cpu_t* cpu, int32_t first_count_ptr, int32_t next_count_ptr);
void syscall_Bkey_Set_RepeatTime_Default(cpu_t* cpu);