#include "../headers/syscalls/malloc.h"

// Allocate memory and return the address of the allocated memory
// The data is cleared (from what I observed), so syscall Calloc 
// redirects to this function too and `clear_data` is ignored
void syscall_Malloc(cpu_t* cpu, uint32_t size, uint8_t clear_data) {
    malloc_manager_t* m = cpu->malloc_manager;

    // Use the first 4 bytes to store the size of the malloc
    *(uint32_t*)(&m->mallocs[m->currentSize]) = size;

    R0 = MALLOC_ADDRESS + m->currentSize + 4; 

    // Clear the data
    for (int i = 0; i < size; i++) {
        m->mallocs[m->currentSize + i + 4] = 0;
    }

    m->currentSize += size + 4;
    m->mallocCount++;

    printf("Run syscall: %salloc #%d (size: %d, addr: 0x%X) Total allocated: %d\n", clear_data ? clear_data == 1 ? "C" : "Re" : "M", m->mallocCount, size, R0, m->currentSize - m->mallocCount*4);
}

void syscall_Realloc(cpu_t* cpu, uint32_t ptr, uint32_t size) {
    uint8_t* mallocs = cpu->malloc_manager->mallocs;

    syscall_Malloc(cpu, size, 2);

    if (ptr == 0x0) return; // No pointer to reallocate, act as a malloc only

    // Get the size of the previous malloc
    uint32_t prev_size = *(uint32_t*)(&mallocs[ptr - MALLOC_ADDRESS - 4]);

    // Transfer data to the new malloc
    for (int i = 0; i < prev_size; i++) {
        mallocs[R0 - MALLOC_ADDRESS + i] = mallocs[ptr - MALLOC_ADDRESS + i];
    }

    syscall_Free(cpu, ptr);
}

void syscall_Free(cpu_t* cpu, uint32_t ptr) {
    cpu->malloc_manager->mallocCount--;

    // printf("Run syscall: Free #%d (addr: 0x%X) Total allocated: %d\n", cpu->malloc_manager->mallocCount, ptr, cpu->malloc_manager->currentSize - cpu->malloc_manager->mallocCount);
}