#include "../headers/syscalls/misc.h"

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_AppName.HTM
// Copies the registered name for the running application into the character array dest. 
// dest must be able to hold 9 bytes. dest is returned.
void syscall_GetAppName(cpu_t* cpu, uint32_t dest_ptr) {
    for (int i = 0; i < 9; i++) {
        mem_write(cpu, dest_ptr + i, cpu->mem->rom[0x20 + i], 1);
    }

    // printf("Run syscall: GetAppName (%s)\n", (const char*)get_memory_for_address(cpu, dest_ptr));

    R0 = dest_ptr; // Return the buffer
}

// Returns the OS version as ASCIIZ string in version.
// Format: "MM.mm.ssss", f. i. "01.05.0000". Space for 10 characters is required.
void syscall_System_GetOSVersion(cpu_t* cpu, uint32_t version_ptr) {
    const char* version = "02.09.0000";

    for (int i = 0; i < 10; i++) {
        mem_write(cpu, version_ptr + i, version[i], 1);
    }

    // printf("Run syscall: System_GetOSVersion (%s)\n", version);
}

void syscall_GetVRAMAddress(cpu_t* cpu) {
    R0 = VRAM_ADDRESS;
    // printf("Run syscall: GetVRAMAddress (-> 0x%08X)\n", R0);
}

void syscall_GlibGetAddinLibInf(cpu_t* cpu, uint32_t a_ptr, uint32_t b_ptr, uint32_t c_ptr) {
    // Mimic Casio SDK Emulator
    mem_write(cpu, a_ptr, 0x0, 4);
    mem_write(cpu, b_ptr, 0x1, 4);
    mem_write(cpu, c_ptr, 0x1, 4);

    cpu->r[0] = 0x1;// 0x0;
    cpu->r[2] = 0xA0151F28;
    cpu->r[3] = 0x0;
    cpu->r[4] = 0x1;

    printf("Run syscall: GlibGetAddinLibInf\n");
}

void syscall_GlibGetOSVersionInfo(cpu_t* cpu, uint32_t a_ptr, uint32_t b_ptr, uint32_t c_ptr, uint32_t d_ptr) {
    // Mimic Casio SDK Emulator
    mem_write(cpu, a_ptr, 0x1, 1);
    mem_write(cpu, b_ptr, 0x3, 1);
    mem_write(cpu, c_ptr, 0x0, 2);
    mem_write(cpu, d_ptr, 0x0, 2);
    
    cpu->r[0] = 0x1;
    cpu->r[2] = 0x3;
    cpu->r[3] = 0x1;
    cpu->r[4] = 0x0;

    // printf("Run syscall: GlibGetOSVersionInfo\n");
}

// Returns number of 1/128 seconds since an arbitrary date.
void syscall_RTC_GetTicks(cpu_t* cpu) {
    R0 = cpu->RTC_Ticks;
    // printf("Run syscall: RTC_GetTicks %d\n", R0);
}

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_Expressions.HTM
// Returns a display string in string according to opcode.
// The string is copied. The size of the target buffer string must be 32 bytes.
// Valid ranges : [0-254], [32512-32756], [58624-59104], [63232-63452], [63745-63778]
void syscall_OpcodeToStr(cpu_t* cpu, uint32_t opcode, uint32_t string_ptr) {
    if (opcode < 0 || opcode > 63778 || (opcode > 254 && opcode < 32512) || (opcode > 32756 && opcode < 58624) || 
       (opcode > 59104 && opcode < 63232) || (opcode > 63452 && opcode < 63745)) {
        printf("[Warning] In syscall OpcodeToStr: Invalid opcode: %d\n", opcode);
        return;
    }
    if (opcode >= 63745) opcode -= 63745 - 1202;
    else if (opcode >= 63232) opcode -= 63232 - 981;
    else if (opcode >= 58624) opcode -= 58624 - 500;
    else if (opcode >= 32512) opcode -= 32512 - 255;

    for (int i = 0; i <= 15; i++) {
        char current = cpu->disp->opcodes[opcode * 15 + i];

        if (current == '\0' || i == 15) {
            mem_write(cpu, string_ptr + i, '\0', 1);
            break;
        }

        mem_write(cpu, string_ptr + i, current, 1);
    }

    // printf("Run syscall: OpcodeToStr (0x%X) => %s\n", opcode, (const char*)get_memory_for_address(cpu, string_ptr));

    R0 = string_ptr;
}

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_Sleep.HTM
void syscall_OS_inner_Sleep(cpu_t* cpu, uint32_t ms) {
    // printf("Run syscall: OS_inner_Sleep %d\n", ms);
#ifdef USE_EMSCRIPT
    emscripten_sleep(ms * .75);
#endif
}