#include "../headers/syscalls/syscalls.h"

// Syscall entry point
void run_syscall(cpu_t* cpu) {
    uint32_t syscall_code = R0;

    if (syscall_code == 0x135)      syscall_GetVRAMAddress(cpu);
    else if (syscall_code == 0x143) syscall_Bdisp_AllClr_VRAM(cpu);
    else if (syscall_code == 0x144) syscall_Bdisp_AllClr_DDVRAM(cpu);
    else if (syscall_code == 0x028) syscall_Bdisp_PutDisp_DD(cpu);
    else if (syscall_code == 0x142) syscall_Bdisp_AllClr_DD(cpu);
    else if (syscall_code == 0x145) syscall_Bdisp_GetDisp_VRAM(cpu, R4);
    else if (syscall_code == 0x146) syscall_Bdisp_SetPoint_VRAM(cpu, R4, R5, R6);
    else if (syscall_code == 0x030) syscall_Bdisp_DrawLine_VRAM(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0x031) syscall_Bdisp_ClearLine_VRAM(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0x14D) syscall_Bdisp_AreaReverseVRAM(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0x01C) syscall_Bdisp_WriteGraph_VRAM(cpu, R4);
    
    else if (syscall_code == 0x813) syscall_SaveDisp(cpu, R4);
    else if (syscall_code == 0x814) syscall_RestoreDisp(cpu, R4);
    else if (syscall_code == 0x8FE) syscall_PopupWin(cpu, R4);

    // https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_AUX.htm
    else if (syscall_code == 0x4E6) syscall_HourGlass(cpu);
    else if (syscall_code == 0x268) syscall_GetFKeyIconPointer(cpu, R4, R5);
    else if (syscall_code == 0x4D1) syscall_DisplayFKeyIcon(cpu, R4, R5);

    else if (syscall_code == 0x807) syscall_Locate(cpu, R4, R5);
    else if (syscall_code == 0x808) syscall_Print(cpu, R4);
    else if (syscall_code == 0x809) syscall_PrintRev(cpu, R4);
    else if (syscall_code == 0x80A) syscall_PrintC(cpu, R4);
    else if (syscall_code == 0x80B) syscall_PrintRevC(cpu, R4);
    else if (syscall_code == 0x80C) syscall_PrintLine(cpu, R4, R5);
    else if (syscall_code == 0x9AD) syscall_PrintXY(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0xC4F) syscall_PrintMini(cpu, R4, R5, R6, R7);
    
    else if (syscall_code == 0x138) syscall_Cursor_SetPosition(cpu, R4, R5);
    else if (syscall_code == 0x80E) syscall_Cursor_GetFlashStyle(cpu);
    else if (syscall_code == 0x811) syscall_Cursor_SetFlashOn(cpu, R4);
    else if (syscall_code == 0x812) syscall_Cursor_SetFlashOff(cpu);
    else if (syscall_code == 0x13A) syscall_Cursor_SetFlashMode(cpu, R4);

    else if (syscall_code == 0xACD) syscall_Malloc(cpu, R4, 0);
    else if (syscall_code == 0xE6B) syscall_Malloc(cpu, R4, 1); // Calloc
    else if (syscall_code == 0xE6D) syscall_Realloc(cpu, R4, R5);
    else if (syscall_code == 0xACC) syscall_Free(cpu, R4);
    
    else if (syscall_code == 0x462) syscall_GetAppName(cpu, R4);
    else if (syscall_code == 0x2EE) syscall_System_GetOSVersion(cpu, R4);
    else if (syscall_code == 0x014) syscall_GlibGetAddinLibInf(cpu, R4, R5, R6);
    else if (syscall_code == 0x015) syscall_GlibGetOSVersionInfo(cpu, R4, R5, R6, R7);

    else if (syscall_code == 0x90F) syscall_GetKey(cpu, R4);
    else if (syscall_code == 0x247) syscall_KeyBoard_GetKeyWait(cpu, R4, R5, R6, R7, mem_read(cpu, R15, 4), mem_read(cpu, R15+4, 4));
    else if (syscall_code == 0x241) syscall_Keyboard_ClrBuffer(cpu);
    else if (syscall_code == 0x242) syscall_Bkey_Set_RepeatTime(cpu, R4, R5);
    else if (syscall_code == 0x243) syscall_Bkey_Get_RepeatTime(cpu, R4, R5);
    else if (syscall_code == 0x244) syscall_Bkey_Set_RepeatTime_Default(cpu);
    else if (syscall_code == 0x910) syscall_PutKey(cpu, R4, R5);
    else if (syscall_code == 0x7FC) syscall_OpcodeToStr(cpu, R4, R5);

    else if (syscall_code == 0x03B) syscall_RTC_GetTicks(cpu);
    else if (syscall_code == 0x420) syscall_OS_inner_Sleep(cpu, R4);

    else if (syscall_code == 0x434) syscall_Bfile_CreateEntry(cpu, R4, R5, R6);
    else if (syscall_code == 0x439) syscall_Bfile_DeleteEntry(cpu, R4, R5);
    else if (syscall_code == 0x42C) syscall_Bfile_OpenFile(cpu, R4, R5, R6);    
    else if (syscall_code == 0x42D) syscall_Bfile_CloseFile(cpu, R4);
    else if (syscall_code == 0x435) syscall_Bfile_WriteFile(cpu, R4, R5, R6);
    else if (syscall_code == 0x432) syscall_Bfile_ReadFile(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0x431) syscall_Bfile_SeekFile(cpu, R4, R5);
    else if (syscall_code == 0x42F) syscall_Bfile_GetFileSize(cpu, R4);
    else if (syscall_code == 0x43B) syscall_Bfile_FindFirst(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0x43C) syscall_Bfile_FindNext(cpu, R4, R5, R6);
    else if (syscall_code == 0x43D) syscall_Bfile_FindClose(cpu, R4);
    else if (syscall_code == 0x42E) syscall_Bfile_Bfile_GetMediaFree(cpu, R4, R5);
    
    else if (syscall_code == 0x840) syscall_MCSGetDlen2(cpu, R4, R5, R6);
    else if (syscall_code == 0x841) syscall_MCSGetData1(cpu, R4, R5, R6);
    else if (syscall_code == 0x82B) syscall_MCSPutVar2(cpu, R4, R5, R6, R7);
    else if (syscall_code == 0x830) syscall_MCSOvwDat2(cpu, R4, R5, R6, R7, mem_read(cpu, R15, 4));
    else if (syscall_code == 0x836) syscall_MCSDelVar2(cpu, R4, R5);
    else if (syscall_code == 0x367) syscall_MCS_DeleteItem(cpu, R4, R5);
    else if (syscall_code == 0x368) syscall_MCS_GetState(cpu, R4, R5, R6);
    else if (syscall_code == 0x371) syscall_MCS_OverwriteData(cpu, R4, R5, R6, R7, mem_read(cpu, R15, 4));
    else if (syscall_code == 0x376) syscall_MCS_SearchDirectoryItem(cpu, R4, R5, R6, R7, mem_read(cpu, R15, 4), mem_read(cpu, R15+4, 4));
    
    else if (syscall_code == 0x24C) printf("Skipped syscall: Keyboard_IsSpecialKeyDown\n");
    else if (syscall_code == 0x477); // printf("Skipped syscall: EnableGetkeyToMainFunctionReturn\n");
    else if (syscall_code == 0x478); // printf("Skipped syscall: DisableGetkeyToMainFunctionReturn\n");
    else if (syscall_code == 0x3ED) { // Interrupt_SetOrClrStatusFlags
        // https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_INTERRUPT.HTM
        printf("[Error] Blocking syscall not implemented: Interrupt_SetOrClrStatusFlags 0x%X\n", R5);
        cpu->execution_finished = 1;
    }
    // Can be ignored
    else if (syscall_code == 0x494) printf("Ignored syscall: SetQuitHandler\n");
    else if (syscall_code == 0x013) printf("Ignored syscall: GlibAddinAplExecutionCheck\n");
    else if (syscall_code == 0x3FA) printf("Ignored syscall: Hmem_SetMMU\n");
    // Unknown syscall
    else {
        printf("[Warning] Syscall not implemented, skipping: 0x%03X\n", syscall_code);
    }

    cpu->pc = cpu->pr;
}